/**
 * A simple 1st-order allpass filter class.
 */
class _FV3_(allpass)
{
 public:
  _FV3_(allpass)();
  virtual _FV3_(~allpass)();
  void free();
  void setsize(long size) throw(std::bad_alloc);
  long getsize();
  void mute();
  void         setfeedback(_fv3_float_t val);
  _fv3_float_t getfeedback();
  void         setdecay(_fv3_float_t val);
  _fv3_float_t getdecay();

  /**
   * Retrive the last signal of the delayline.
   * @return The last signal (z^(-bufsize)) of the delayline. This is the same as get_z(getsize()).
   */
  inline _fv3_float_t getlast()
  {
    if(bufsize == 0) return 0;
    return buffer[bufidx];
  }
  
  /**
   * Retrive the signal of the delayline.
   * @param[in] index specifies the index of the delayline.
   * @return The signal value (z^(-index)).
   */
  inline _fv3_float_t get_z(long index)
  {
    if(bufsize == 0) return 0;
    if(index > bufsize) index = bufsize;
    if(index <= 0) index = 1;
    long readpoint = bufidx - index;
    if(readpoint < 0) readpoint += bufsize;
    return buffer[readpoint];
  }

  /**
   * A simple normal allpass filter process method.
   * @param[in] input signal.
   */  
  inline _fv3_float_t process(_fv3_float_t input)
  {
    if(bufsize == 0) return input;
    _fv3_float_t buffer_tmp = buffer[bufidx];
    input += feedback * buffer_tmp;
    _fv3_float_t output = buffer_tmp - feedback * input; UNDENORMAL(output);
    buffer[bufidx] = input;
    bufidx ++; if(bufidx >= bufsize) bufidx = 0;
    return output;
  }
  inline _fv3_float_t operator()(_fv3_float_t input){return this->process(input);}

  /**
   * An allpass filter with decay.
   * @param[in] input signal.
   */  
  inline _fv3_float_t process_dc(_fv3_float_t input)
  {
    if(bufsize == 0) return input;
    _fv3_float_t buffer_tmp = buffer[bufidx];
    input += feedback * buffer_tmp;
    _fv3_float_t output = decay * buffer_tmp - feedback * input; UNDENORMAL(output);
    buffer[bufidx] = input;
    bufidx ++; if(bufidx >= bufsize) bufidx = 0;
    return output;
  }

  /**
   * An original freeverb version of allpass filter.
   * The true allpass filter is obtained if g = (sqrt(5)-1)/2 ~ 0.618 (reciprocal of the golden ratio)
   * while that of freeverb is 0.5.
   * @param[in] input signal.
   */
  inline _fv3_float_t process_ov(_fv3_float_t input)
  {
    if(bufsize == 0) return input;
    _fv3_float_t bufout = buffer[bufidx];
    UNDENORMAL(bufout);
    buffer[bufidx] = input + (bufout * feedback);
    bufidx ++; if(bufidx >= bufsize) bufidx = 0;
    return (bufout - input);
  }

 private:
  _FV3_(allpass)(const _FV3_(allpass)& x);
  _FV3_(allpass)& operator=(const _FV3_(allpass)& x);
  _fv3_float_t feedback, *buffer, decay;
  long bufsize, bufidx;
};

/**
 * A 1st-order allpass filter with a allpass interplated delay modulation class.
 */
class _FV3_(allpassm)
{
 public:
  _FV3_(allpassm)();
  virtual _FV3_(~allpassm)();
  void free();
  void setsize(long size) throw(std::bad_alloc);
  void setsize(long size, long modsize) throw(std::bad_alloc);
  long getsize();
  long getdelaysize();
  long getmodulationsize();
  void mute();
  void         setfeedback(_fv3_float_t val);
  _fv3_float_t getfeedback();
  void         setdecay(_fv3_float_t val);
  _fv3_float_t getdecay();
  void         set_90degfq(_fv3_float_t fc, _fv3_float_t fs);

  /**
   * An allpass filter with a decay.
   * @param[in] input signal.
   * @return processed signal.
   */  
  inline _fv3_float_t process(_fv3_float_t input){return process(input, 0);}
  inline _fv3_float_t operator()(_fv3_float_t input){return this->process(input);}

  /**
   * An allpass filter with a modulation and a decay.
   * @param[in] input signal.
   * @return processed signal.
   */  
  inline _fv3_float_t process(_fv3_float_t input, _fv3_float_t modulation)
  {
    if(bufsize == 0) return input;

    if(modulation < -1) modulation = -1; if(modulation > 1) modulation = 1;
    modulation += 1.;
    modulation *= (_fv3_float_t)modulationsize;
    _fv3_float_t floor_mod = std::floor(modulation); // >= 0
    _fv3_float_t frac = modulation - floor_mod; // >= 0
    
    long readidx_a = readidx - (long)floor_mod;
    if(readidx_a < 0) readidx_a += bufsize;
    long readidx_b = readidx_a - 1;
    if(readidx_b < 0) readidx_b += bufsize;
    
    _fv3_float_t temp = buffer[readidx_b] + buffer[readidx_a]*(1-frac) - (1-frac)*z_1;
    UNDENORMAL(temp);
    z_1 = temp;
    
    readidx ++; if(readidx >= bufsize) readidx = 0;

    buffer[writeidx] = input + temp * feedback;
    _fv3_float_t output = decay * temp - buffer[writeidx] * feedback;
    writeidx ++; if(writeidx >= bufsize) writeidx = 0;
    
    return output;
  }
  inline _fv3_float_t operator()(_fv3_float_t input, _fv3_float_t modulation){return this->process(input,modulation);}
 
 private:
  _FV3_(allpassm)(const _FV3_(allpassm)& x);
  _FV3_(allpassm)& operator=(const _FV3_(allpassm)& x);
  _fv3_float_t feedback, *buffer, z_1, decay;
  long bufsize, readidx, writeidx, delaysize, modulationsize;
};

/**
 * A simple 2nd-order allpass filter class.
 */
class _FV3_(allpass2)
{
 public:
  _FV3_(allpass2)();
  virtual _FV3_(~allpass2)();
  void free();

  /**
   * set allpass delay size.
   * @param[in] size1 The inner allpass delay size.
   * @param[in] size2 The outer allpass delay size.
   */
  void setsize(long size1, long size2) throw(std::bad_alloc);
  
  long getsize1();
  long getsize2();

  /**
   * set the innter allpass feedback.
   * @param[in] val The innter allpass feedback.
   */
  void setfeedback1(_fv3_float_t val);
  
  _fv3_float_t getfeedback1();
  
  /**
   * set the outter allpass feedback.
   * @param[in] val The outter allpass feedback.
   */
  void setfeedback2(_fv3_float_t val);
  
  _fv3_float_t getfeedback2();

  /**
   * set the innter allpass decay.
   * @param[in] val The innter allpass decay.
   */
  void setdecay1(_fv3_float_t val);
  
  _fv3_float_t getdecay1();
  
  /**
   * set the outter allpass decay.
   * @param[in] val The outter allpass decay.
   */
  void setdecay2(_fv3_float_t val);
  
  _fv3_float_t getdecay2();
  
  /**
   * Retrive the last signal of the inner delayline.
   * @return The last signal (z^(-bufsize)) of the inner delayline.
   */
  inline _fv3_float_t getlast1(){return buffer1[bufidx1];}

  /**
   * Retrive the last signal of the outter delayline.
   * @return The last signal (z^(-bufsize)) of the outter delayline.
   */
  inline _fv3_float_t getlast2(){return buffer2[bufidx2];}

  /**
   * A 2nd order simple allpass filter with decay.
   * @param[in] input signal.
   * @return processed signal.
   */  
  inline _fv3_float_t process(_fv3_float_t input)
  {
    if(bufsize1 == 0||bufsize2 == 0) return input;

    input += feedback2*buffer2[bufidx2];
    _fv3_float_t output = decay2 * buffer2[bufidx2] - input*feedback2; UNDENORMAL(output);

    input += feedback1*buffer1[bufidx1];
    buffer2[bufidx2] = decay1 * buffer1[bufidx1] - input*feedback1; UNDENORMAL(buffer2[bufidx2]);

    buffer1[bufidx1] = input;

    bufidx1 ++; if(bufidx1 >= bufsize1) bufidx1 = 0;
    bufidx2 ++; if(bufidx2 >= bufsize2) bufidx2 = 0;
    return output;
  }
  inline _fv3_float_t operator()(_fv3_float_t input){return this->process(input);}
  
  void mute();
 
 private:
  _FV3_(allpass2)(const _FV3_(allpass2)& x);
  _FV3_(allpass2)& operator=(const _FV3_(allpass2)& x);
  _fv3_float_t feedback1, feedback2, decay1, decay2, *buffer1, *buffer2;
  long bufsize1, bufidx1, bufsize2, bufidx2;
};

/**
 * A simple 3rd-order allpass filter with internal comb filter class.
 */
class _FV3_(allpass3c)
{
 public:
  _FV3_(allpass3c)();
  virtual _FV3_(~allpass3c)();
  void free();

  void setsize(long size1, long size2, long size3) throw(std::bad_alloc);
  long getsize1();
  long getsize2();
  long getsize3();
  
  inline _fv3_float_t getlast1(){return buffer1[bufidx1];}
  inline _fv3_float_t getlast2(){return buffer2[bufidx2];}
  inline _fv3_float_t getlast3(){return buffer3[bufidx3];}

  inline _fv3_float_t process(_fv3_float_t input)
  {
    if(bufsize1 == 0||bufsize2 == 0||bufsize3 == 0) return input;

    input += feedback3 * buffer3[bufidx3];
    _fv3_float_t output = decay3 * buffer3[bufidx3] - feedback3 * input; UNDENORMAL(output);

    input += feedback2 * buffer2[bufidx2];
    buffer3[bufidx3] = decay2 * buffer2[bufidx2] - feedback2 * input; UNDENORMAL(buffer3[bufidx3]);

    input += feedback1 * buffer1[bufidx1];
    buffer2[bufidx2] = decay1 * buffer1[bufidx1] - feedback1 * input; UNDENORMAL(buffer2[bufidx2]);

    buffer1[bufidx1] = this->iir_lpf_in.process(input);

    bufidx1 ++; if(bufidx1 >= bufsize1) bufidx1 = 0;
    bufidx2 ++; if(bufidx2 >= bufsize2) bufidx2 = 0;
    bufidx3 ++; if(bufidx3 >= bufsize3) bufidx3 = 0;
    return output;
  }
  inline _fv3_float_t operator()(_fv3_float_t input){return this->process(input);}

  void mute();
  void setfeedback1(_fv3_float_t val);
  void setfeedback2(_fv3_float_t val);
  void setfeedback3(_fv3_float_t val);
  _fv3_float_t getfeedback1();
  _fv3_float_t getfeedback2();
  _fv3_float_t getfeedback3();

  void setdecay1(_fv3_float_t val);
  void setdecay2(_fv3_float_t val);
  void setdecay3(_fv3_float_t val);
  _fv3_float_t getdecay1();
  _fv3_float_t getdecay2();
  _fv3_float_t getdecay3();
  
  void setLPF(_fv3_float_t fc, _fv3_float_t fs){this->iir_lpf_in.setLPFA(fc,fs);}

 private:
  _FV3_(allpass3c)(const _FV3_(allpass3c)& x);
  _FV3_(allpass3c)& operator=(const _FV3_(allpass3c)& x);
  _fv3_float_t feedback1, feedback2, feedback3, *buffer1, *buffer2, *buffer3, decay1, decay2, decay3;
  long bufsize1, bufidx1, bufsize2, bufidx2, bufsize3, bufidx3;
  _FV3_(iir_1st) iir_lpf_in;
};
